// ByteMatrix.cpp: implementation of the ByteMatrix class.
//
//////////////////////////////////////////////////////////////////////

#include "ByteMatrix.h"
#include "Utilities.h"
#include "gioptr.h"

// GridIOLibrary global object
extern GridIOLibrary gridio;

#include <stdio.h>

#include <scew.h>

#ifdef __GARPWIN
#include <direct.h>
#endif

// ============================================================================
//  Virtual ByteMatrix class implementation
// ============================================================================
ByteMatrix::ByteMatrix() {}

// ============================================================================
ByteMatrix::~ByteMatrix() {}

// ============================================================================
char * ByteMatrix::toXML()
{
  const int iStringLen = 1024;
  char * strXML;

  strXML = new char[iStringLen];

  // create the tag
  sprintf(strXML, "<ByteMatrix Type=\"%s\"/>", type());

  if (strlen(strXML) > iStringLen)
    throw GarpException(82, "String size exceeded in ByteMatrix::toXML()");

  return strXML;
}
// ============================================================================
ByteMatrix * ByteMatrix::fromXML(scew_element const * pXMLMatrix)
{
  ByteMatrix * newMatrix;
  char * strValue;

  if (pXMLMatrix)
  {
    strValue = GarpUtil::getXMLAttrAsString(pXMLMatrix, "Type");

    if (!strValue)
      throw GarpException(47, "ByteMatrix has no Type attribute");

    // alocate matrix of the right type
    if      (strcmp(strValue, "RawByteMatrixInDisk") == 0)   newMatrix = new RawByteMatrixInDisk;
    else if (strcmp(strValue, "RawByteMatrixInMemory") == 0) newMatrix = new RawByteMatrixInMemory;
    else if (strcmp(strValue, "BooleanByteMatrix") == 0)     newMatrix = new BooleanByteMatrix;

#ifdef __GARPWIN
    else if (strcmp(strValue, "BitmapByteMatrix") == 0)      newMatrix = new BitmapByteMatrix;
#endif

#ifdef __GRIDIO
    else if (strcmp(strValue, "ArcGridByteMatrix") == 0)     newMatrix = new ArcGridByteMatrix;
#endif
    else throw GarpException(35, "Matrix type not supported");

    // get filename
    strValue = GarpUtil::getXMLAttrAsString(pXMLMatrix, "Filename");

    if (!strValue)
      throw GarpException(48, "ByteMatrix has no Filename attribute");

    newMatrix->open(strValue);

    return newMatrix;
  }
  else
    throw(1004, "Cannot read attributes of ByteMatrix. XML node is empty.");

  return NULL;
}

// ============================================================================
//  RawByteMatrixInDisk class implementation
// ============================================================================
RawByteMatrixInDisk::RawByteMatrixInDisk()
{
  file = NULL;
  strcpy(strFilename, "");
  intCursor = 0;
}

// ============================================================================
RawByteMatrixInDisk::~RawByteMatrixInDisk()
{
  if (file)
    close();
}

// ============================================================================
char * RawByteMatrixInDisk::toXML()
{
  const int iStringLen = 1024;
  char * strXML;

  strXML = new char[iStringLen];

  // create the tag
  sprintf(strXML, "<ByteMatrix Type=\"%s\" Filename=\"%s\"/>", type(), strFilename);

  if (strlen(strXML) > iStringLen)
    throw GarpException(82, "String size exceeded in RawByteMatrixInDisk::toXML()");

  return strXML;
}

// ============================================================================
void RawByteMatrixInDisk::rewind()
{
  fseek(file, 0, 0);
  intCursor = 0;
}

// ============================================================================
void RawByteMatrixInDisk::set(long pos, BYTE val)
{
  // check if it is necessary to make a seek operation
  if (pos != intCursor)
    fseek(file, pos, 0);

  // set value
  fputc((BYTE) val, file);
  intCursor = pos + 1;
}

// ============================================================================
BYTE RawByteMatrixInDisk::get(long pos)
{
  BYTE val;

  // check if it is necessary to make a seek operation
  if (pos != intCursor + 1)
    fseek(file, pos, 0);

  // get value
  val = fgetc(file);
  intCursor = pos + 1;

  return val;
}

// ============================================================================
void RawByteMatrixInDisk::create(char * strParamFilename, long intFileSize)
{
  // set object property
  strcpy(strFilename, strParamFilename);

  // create a new file
  file = fopen(strFilename, "wb");

  if (file)
  {
    BYTE buffer[FILE_CHUNK + 1];
    memset((void *) buffer, (BYTE) 0, sizeof(BYTE) * FILE_CHUNK + 1);

    int size = 0;
    int left = intFileSize;
    while (left)
    {
      size += fwrite(buffer, sizeof(BYTE), (left > FILE_CHUNK)? FILE_CHUNK : left, file);
      left = intFileSize - size;
    }

    fclose(file);
  }

  else
  {
    // file not found
    char msg[256];
    sprintf(msg, "Matrix file <%s> could not be opened for writing", strFilename);
    throw GarpException(1, msg);
  }
}
// ============================================================================
void RawByteMatrixInDisk::open(char * strParamFilename)
{
  // set object property
  strcpy(strFilename, strParamFilename);

  // try opening it R/W
  file = fopen(strFilename, "rb+");

  // then try R/O
  if (!file) file = fopen(strFilename, "rb");

  if (file)
  {
    intCursor = 0;
  }

  else
  {
    // file not found
    char msg[256];
    sprintf(msg, "Matrix file <%s> not found", strFilename);
    throw GarpException(1, msg);
  }

}

// ============================================================================
void RawByteMatrixInDisk::close()
{
  if (file)
  {
    fclose(file);
    file = NULL;
  }
  else
    throw GarpException(1, "Could not close matrix file");
}

// ============================================================================
bool RawByteMatrixInDisk::isOpen()
{
  return (file != NULL);
}

// ============================================================================
ByteMatrix * RawByteMatrixInDisk::clone()
{
  RawByteMatrixInDisk * objClone = new RawByteMatrixInDisk();

  strcpy(objClone->strFilename, this->strFilename);

  return objClone;
}

// ============================================================================

// ============================================================================
//  RawByteMatrixInMemory class implementation
// ============================================================================
RawByteMatrixInMemory::RawByteMatrixInMemory()
{
  saved = true;
  array = NULL;
}

// ============================================================================
RawByteMatrixInMemory::~RawByteMatrixInMemory()
{
  if (isOpen())
    close();

  if (array)
    delete[] array;
}

// ============================================================================
void RawByteMatrixInMemory::set(long pos, BYTE val)
{
  saved &= (array[pos] == val);
  array[pos] = val;
}

// ============================================================================
BYTE RawByteMatrixInMemory::get(long pos)
{
  return array[pos];
}

// ============================================================================
void RawByteMatrixInMemory::create(char * strParamFilename, long intFileSize)
{
  // set object property
  strcpy(strFilename, strParamFilename);

  // save file size;
  intSize = intFileSize;
}
// ============================================================================
void RawByteMatrixInMemory::open(char * strFilename)
{
  if ((strcmp(strFilename, "") == 0) || (strcmp(strFilename, "NULL") == 0))
  {
    // file doesn't exist
    // set array with blank bytes
    array = new BYTE[intSize];
    memset((void *)array, 0, intSize * sizeof(BYTE));
  }

  else
  {
    // open the file
    RawByteMatrixInDisk::open(strFilename);

    // get the file size
    intSize = (int) getFileSize(strFilename);

    array = new BYTE[intSize];

    int intLastPosRead;
    int size = 0;
    while (!feof(file))
    {
      intLastPosRead = size;
      size += fread(array + intLastPosRead, sizeof(BYTE), FILE_CHUNK, file);
    }
  }
}

// ============================================================================
void RawByteMatrixInMemory::save()
{
  if ((strcmp(strFilename, "NULL") == 0) ||
    (strcmp(strFilename, "") == 0))
    throw GarpException(2, "Cannot save dataset with an empty filename");

  if (!saved)
  {
    // get the file size
    int intFileSize = GarpUtil::getFileSize(strFilename);

    // save dump file back to disk
    rewind();
    int size = 0;
    int left = intFileSize;
    while (left > 0)
    {
      size += fwrite(&array[size], sizeof(BYTE), (left > FILE_CHUNK)? FILE_CHUNK : left, file);
      left = intFileSize - size;
    }
  }

  saved = true;
}

// ============================================================================
void RawByteMatrixInMemory::close()
{
  save();
  RawByteMatrixInDisk::close();

  delete[] array;
  array = NULL;
}

// ============================================================================
BYTE * RawByteMatrixInMemory::getArrayPtr()
{
  return array;
}

// ============================================================================
int RawByteMatrixInMemory::size()
{
  return intSize;
}

// ============================================================================

// ============================================================================
//  RawByteMatrixInMMF class implementation
// ============================================================================

#ifdef __GARPWIN
// ============================================================================
RawByteMatrixInMMFile::RawByteMatrixInMMFile()
{
  iSize = 0;
  strcpy(sFilename, "");
  hFile = NULL;
  hMMFile = NULL;
  bArray = NULL;
}

// ============================================================================
RawByteMatrixInMMFile::~RawByteMatrixInMMFile()
{
  if (bArray)
    close();
}

// ============================================================================
ByteMatrix * RawByteMatrixInMMFile::clone()
{
  return NULL;
}

// ============================================================================
void RawByteMatrixInMMFile::set(long pos, BYTE val)
{
  throw GarpException(1, "Memory-mapped byte matrixes are read-only.");
}

// ============================================================================
BYTE RawByteMatrixInMMFile::get(long pos)
{
  return bArray[pos];
}

// ============================================================================
void RawByteMatrixInMMFile::create(char * strParamFilename, long intFileSize)
{
  // set object property
  strcpy(sFilename, strParamFilename);

  // save file size;
  iSize = intFileSize;
}
// ============================================================================
void RawByteMatrixInMMFile::open(char * strFilename)
{
  char * strSlash, strFullFilename[1024], strPath[1024];

  //bArray = new BYTE[iSize];
  //memset((void *)bArray, 0, iSize * sizeof(BYTE));

  _getcwd(strPath, 1024);
  sprintf(strFullFilename, "%s\\%s", strPath, strFilename);

  // "\" is not a valid MMFile name so, replace all \ by +
  strSlash = strFullFilename;
  while (strSlash = strchr(strSlash, '\\'))
    strSlash[0] = '/';

  // try to open the memory-mapped file
  hMMFile = OpenFileMapping(FILE_MAP_READ, FALSE, strFullFilename);

  if (!hMMFile)
  {
    // memory-mapped file has not been created: so create it
    hFile = CreateFile((LPCSTR) strFilename, GENERIC_READ, FILE_SHARE_READ, NULL,
                                 OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

    if (!hFile)
      throw GarpException(1, "Could not open memory-maped file.");

    hMMFile = CreateFileMapping(hFile, NULL, PAGE_READONLY, 0, iSize, strFullFilename);

    if (!hMMFile)
      throw GarpException(1, "Cannot create memory-mapped file");
  }

  // map a view of the mmf
  bArray = (BYTE *) MapViewOfFile(hMMFile, FILE_MAP_READ, 0, 0, 0);
}

// ============================================================================
void RawByteMatrixInMMFile::close()
{
  UnmapViewOfFile(bArray);
  CloseHandle(hMMFile);
  CloseHandle(hFile);

  hMMFile = NULL;
  hFile = NULL;
  bArray = NULL;
}

// ============================================================================
int RawByteMatrixInMMFile::size()
{
  return iSize;
}

// ============================================================================

// ============================================================================
//  BitmapByteMatrix class implementation
// ============================================================================
BitmapByteMatrix::BitmapByteMatrix()
{
}

// ============================================================================
BitmapByteMatrix::~BitmapByteMatrix()
{
}

// ============================================================================
void BitmapByteMatrix::set(long pos, BYTE val)
{
  // byte array alignment depends on the color depth
  BYTE valR, valG, valB;
  long x1, y1, y2, pos2;
  DWORD offset;

  if (val == 0)
  {
    valR = 64;
    valG = 128;
    valB = 64;
  }

  else if (val == 1)
  {
    valR = 128;
    valG = 255;
    valB = 128;
  }

  else if (val == 255)
  {
    valR = 0;
    valG = 0;
    valB = 0;
  }

  // invert the y axis
  x1 = pos % iColumns;
  y1 = pos / iColumns;
  y2 = iRows - y1 - 1;
  pos2 = x1 + y2 * iColumns;
  offset = pos2 * iPixelWidth;

  if (iColorDepth == 16)
  {
    WORD color = 0;
    // reduce color width to 5 bits
    int red, green, blue;
    blue = valB;
    green = valG;
    red = valR;
    color += blue / 8;
    color += (green / 8) << 5;
    color += (red / 8) << 10;

    array[offset + 0] = LOBYTE(color);
    array[offset + 1] = HIBYTE(color);
  }

  else if (iColorDepth == 24)
  {
    RGBTRIPLE * rgb = (RGBTRIPLE*) &array[offset];
    rgb->rgbtBlue  = valB;
    rgb->rgbtGreen = valG;
    rgb->rgbtRed   = valR;
  }

  else if (iColorDepth == 32)
  {
    RGBTRIPLE * rgb = (RGBTRIPLE*) &array[offset];
    rgb->rgbtBlue  = valB;
    rgb->rgbtGreen = valG;
    rgb->rgbtRed   = valR;
  }

  else
  {
    array[offset] = val;
  }
}

// ============================================================================
BYTE BitmapByteMatrix::get(long pos)
{
  DWORD offset = pos * iPixelWidth;

  return array[offset];
}

// ============================================================================
void BitmapByteMatrix::initialize(int iCDepth, int iCol, int iRw)
{
  iColumns = iCol;
  iRows    = iRw;

  iColorDepth = iCDepth;              // in bits
  iPixelWidth = iColorDepth / 8;		// in bytes
}

// ============================================================================
//  ArcGridByteMatrix class implementation
// ============================================================================
#endif

#ifdef __GRIDIO
// ============================================================================
ArcGridByteMatrix::ArcGridByteMatrix()
{
  strcpy(strGridName, "");
  iType = CELLINT;
  iChannel = -1;
  iEvalType = -1;
  iRows = 0;
  iCols = 0;
  iCursor = -1;
  dCellSize = 0.0;
  oRowBuf = NULL;

  for (int i = 0; i < 4; i++)
    dBndBox[i] = 0.0;
}

// ============================================================================
ArcGridByteMatrix::~ArcGridByteMatrix()
{
}

// ============================================================================
ByteMatrix * ArcGridByteMatrix::clone()
{ throw GarpException(41, "ArcGridByteMatrix::clone() method is not implemented yet."); }

// ============================================================================
void ArcGridByteMatrix::getCoordinates(long pos, double& dX, double& dY)
{
  throw GarpException(1, "Please review these calculations before using this method");

  dX = dBndBox[0] + ((double)(pos % iCols) * dCellSize);
  dY = dBndBox[3] - ((double)(pos / iCols) * dCellSize);
}

// ============================================================================
void ArcGridByteMatrix::getIndexes(long pos, int& iRowAux, int& iColAux)
{
  throw GarpException(1, "Please review these calculations before using this method");

  iColAux = pos % iCols;
  iRowAux = pos / iCols;
}

// ============================================================================
BYTE ArcGridByteMatrix::get(long pos)
{
  int iRes, iColAux, iRowAux, iValue;

  // convert pos into (Row, Col) pair
  getIndexes(pos, iRowAux, iColAux);

  iRes = gridio.GetWindowCellInt(iChannel, iColAux, iRowAux, &iValue);

  return scale((int) iValue);
}

// ============================================================================
void ArcGridByteMatrix::set(long pos, BYTE val)
{
  int iRes, iColAux, iRowAux;

  // convert pos into (Row, Col) pair
  getIndexes(pos, iRowAux, iColAux);

  iRes = gridio.PutWindowCellInt(iChannel, iColAux, iRowAux, (int) unscale(val));
}

// ============================================================================
void ArcGridByteMatrix::create(char * strFilenameAux, double dCellSizeAux,
                 double dXMaxAux, double dYMaxAux,
                 double dXMinAux, double dYMinAux,
                 double dMinValueAux, double dMaxValueAux)
{
  dBndBox[0] = dXMinAux;
  dBndBox[1] = dXMaxAux;
  dBndBox[2] = dYMinAux;
  dBndBox[3] = dYMaxAux;

  dMaxValue = dMaxValueAux;
  dMinValue = dMinValueAux;

  dCellSize = dCellSizeAux;

  dValueCoef = (dMaxValue - dMinValue) / SCALED_VALUE_RANGE;

  strcpy(strGridName, strFilenameAux);
  iChannel = gridio.CellLayerCreate(strGridName, READWRITE, CELLIO, CELLINT, dCellSize, dBndBox);

  if (iChannel < 0)
    throw GarpException(1, "Could not create Arc/Info Grid");
}

// ============================================================================
void ArcGridByteMatrix::open(char * strFilaname)
{
  int iRes;
  double dMeanValue, dStdDevValue;

  iChannel = gridio.CellLayerOpen(strFilaname, READWRITE, CELLIO, &iType, &dCellSize);

  if (iChannel < 0)
    throw GarpException(1, "Could not open Arc/Info Grid");

  if (iType != CELLINT)
    throw GarpException(1, "Float grids are not supported.");

  iRes = gridio.StaGetDbl(strFilaname, &dMinValue, &dMaxValue, &dMeanValue, &dStdDevValue);
  iRes = gridio.BndCellRead(strFilaname, dBndBox);

  dValueCoef = (dMaxValue - dMinValue) / SCALED_VALUE_RANGE;
}

// ================================================================================
BYTE ArcGridByteMatrix::scale(double dUnscaledValue)
{
  BYTE b;
  b = (BYTE) (int)( ( (dUnscaledValue - dMinValue) / dValueCoef ) + MIN_SCALED_VALUE);
  return b;
}

// ================================================================================
double ArcGridByteMatrix::unscale(BYTE bytScaledValue)
{
  double v;
  v = ( ( (double) bytScaledValue - MIN_SCALED_VALUE) * dValueCoef ) + dMinValue;
  return v;
}

// ============================================================================
void ArcGridByteMatrix::close()
{
  gridio.CellLyrClose(iChannel);
}

// ============================================================================
bool ArcGridByteMatrix::isOpen()
{
  return (iChannel >= 0);
}
// ============================================================================
#endif //__GRIDIO

// ============================================================================
//  BooleanByteMatrix class implementation
// ============================================================================
BooleanByteMatrix::BooleanByteMatrix()
{
  intPresencePointCount = 0;

  for (int i = 0; i < MAX_PRESENCE_POINTS; i++)
    intPresencePointPosition[i] = -1;
}

// ============================================================================
BooleanByteMatrix::~BooleanByteMatrix() {}
// ============================================================================
ByteMatrix * BooleanByteMatrix::clone()
{ 
	throw GarpException(41, "BooleanByteMatrix::clone() method is not implemented yet."); 
	return NULL;
}

// ============================================================================
void BooleanByteMatrix::copyPresenceArray(int& intPoints, long * intPositionArray)
{
  for (int i = 0; i < intPresencePointCount; i++)
    intPositionArray[i] = intPresencePointPosition[i];

  intPoints = intPresencePointCount;
}

// ============================================================================
void BooleanByteMatrix::set(long pos, BYTE val)
{
  int i, j;

  // find the position where presence point located at <pos> should be
  i = 0;
  while ((i < intPresencePointCount) && (intPresencePointPosition[i] < pos))
    i++;

  if (val == PRESENCE)
  {
    // try to add point (if it is not there yet)
    if (intPresencePointPosition[i] != pos)
    {
      // check if there is enough space to add one more point
      if (intPresencePointCount >= MAX_PRESENCE_POINTS - 2)
        throw GarpException(42, "Overflow: max number of presence points reached.");

      // <i> is the position it should be inserted
      // move other values down
      for (j = intPresencePointCount - 1; j >= i; j--)
        intPresencePointPosition[j + 1] = intPresencePointPosition[j];

      // add presence point
      intPresencePointPosition[i] = pos;
      intPresencePointCount++;
    }
  }

  else
  {
    // try to remove point from layer if it is there
    if (intPresencePointPosition[i] == pos)
    {
      // <i> is the position it is located
      // move other values down
      for (j = i; j < intPresencePointCount - 1; j++)
        intPresencePointPosition[j] = intPresencePointPosition[j + 1];

      // remove presence point
      intPresencePointCount--;
    }
  }
}

// ============================================================================
BYTE BooleanByteMatrix::get(long pos)
{
  // find the position where presence point located at <pos> should be
  int i = 0;
  while ((i < intPresencePointCount) && (intPresencePointPosition[i] < pos))
    i++;

  // check if there is a presence there
  if (intPresencePointPosition[i] == pos)
    // presence point found
    return PRESENCE;
  else
    // no presence point found
    return ABSENCE;
}

// ============================================================================
void BooleanByteMatrix::rewind() {}

// ============================================================================
void BooleanByteMatrix::open(char * strFilename)
{ throw GarpException(41, "BooleanByteMatrix::open() not implemented yet."); }

// ============================================================================
void BooleanByteMatrix::save()
{ throw GarpException(41, "BooleanByteMatrix::save() not implemented yet."); }

// ============================================================================
void BooleanByteMatrix::close()
{ throw GarpException(41, "BooleanByteMatrix::close() not implemented yet."); }

// ============================================================================
bool BooleanByteMatrix::isOpen()
{ return false; }

// ============================================================================
char * BooleanByteMatrix::toXML()
{ 
	throw GarpException(41, "BooleanByteMatrix::toXML() not implemented yet."); 
	return NULL;
}

// ============================================================================
