/*
 *
 *  Copyright (C) 2001 Numeric Solutions.  All Rights Reserved. 
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  Further, this software is distributed without any warranty that it is
 *  free of the rightful claim of any third person regarding infringement
 *  or the like.  Any license provided herein, whether implied or
 *  otherwise, applies only to this software file.  Patent licenses, if
 *  any, provided herein do not apply to combinations of this program with
 *  other software, or any other product whatsoever.
 * 
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact information: Numeric Solutions support@numericsolutions.com, or:
 * 
 *  http://www.numericsolutions.com
 *
 */

#include <iostream>
#include <stdio.h>
#include <string>
#include <fstream>
#include <vector>
#include <math.h>
#include <assert.h>

using namespace std;

extern string ftostr(float val);
extern string itostr(int val);
extern vector<string> gridDensity(vector<string> ,vector<string>, float radius);
extern vector<string> clipData(vector<string> inData, float xmin, float xmax, 
	float ymin, float ymax);

/*******************************************************************************
 * function that takes a, minimally, x y z file as a vector and clips the x 
 * and y dimension to the input ranges
 ******************************************************************************/
vector<string> clipData(vector<string> inData, float xmin, float xmax, 
	float ymin, float ymax) {
	cerr<<"cliping with: "<<xmin<<" "<<xmax<<" "<<ymin<<" "<<ymax<<endl;
	vector<string> out;
	char const * sLine;
	
	for (int i=0; i<inData.size(); i++) {
		sLine = inData[i].c_str();
		char xstr[20];
		char ystr[20];
		float x;
		float y;
		sscanf(sLine,  "%s %s",  xstr, ystr);
		x = atof(xstr);
		y = atof(ystr);
		if ( x<=xmax && x>=xmin && y<=ymax && y>=ymin) {
			out.push_back(sLine);
		}
	}
	
	return out;
}

/******************************************************************************
 * function that converts a float into a char pointer
 * ***************************************************************************/
string ftostr(float val){
    string retStr;
    char buf[255];
    sprintf(buf, "%f", val);
    retStr = buf;
    return retStr;
}

/*****************************************************************************
 * function convert integer to string
 * ***************************************************************************/
string itostr(int val){
    string retStr;
    char buf[255];
    sprintf(buf, "%i", val);
    retStr = buf;
    return retStr;
}

/******************************************************************************
 * gridDensity -- method to populate a 2D grid based on the density of data 
 * points in an irregularly-spaced data set.
 * 
 * Parameters
 *  gridvec -- a c++ string vector containing strings that can be parsed into 
 * 		x, y, z, row, col and looks like: 
 * 		1214900.000000 620700.000000 300 1 1
 *		1215020.000000 620700.000000 300 2 1
 *		1215140.000000 620700.000000 300 3 1
 *		1215260.000000 620700.000000 300 4 1
 *		1215380.000000 620700.000000 300 5 1
 * 
 * 	dataVec -- a c++ string vector containing strings that can be parsed into 
 * 		x, y, z and looks like:
 *		1214900 620700 357.902130127
 *		1215000 620700 348.8722229004
 *		1215100 620700 337.4031982422
 * 		1215200 620700 313.2538452148
 * 
 *	radius -- the search radius from the center of the grid node
 *
 * Returns
 * 	gridvec -- the populated gridVector object
 *
******************************************************************************/
vector<string> gridDensity(vector<string> gridVec, vector<string> dataVec, 
	float radius) {
	vector<string> out;
	char const * sLine;
	int last = 0;
	int initialSize=gridVec.size();
	
	vector<float> dxvec;
	vector<float> dyvec;
	vector<float> dzvec;
	
	/// create a pre-populated vector with the data x and y's
	for (int ii=0; ii<dataVec.size(); ii++) {
		char const * ssLine;
		//convert to the C lib for use of 
		//string manipulations
		ssLine = dataVec[ii].c_str();
		char px[20];
		char py[20];
		char pp[20];
		sscanf (ssLine,  "%s %s %s",  px, py, pp);
		dxvec.push_back(atof(px));
		dyvec.push_back(atof(py));
		dzvec.push_back(atof(pp));
	}
	
	//iterate through the grid
	for (int i=0; i<gridVec.size(); i++) {
		float percentdone = ((float)i/(float)initialSize)*100;
		int current = (int)percentdone;
		if ( last != current ) {
			cerr<<"\x1b[2J"<<" "<<i<<" "<<initialSize<<" "<<(int)percentdone<<" % done"<<endl;
			last = (int)percentdone;
		}
		
		//convert to the C lib for use of 
		//string manipulations
		sLine = gridVec[i].c_str();
		// get the grid x,y,z's into variables
		char gx[20];
		char gy[20];
		char gp[20];
    char xrow[20];
    char ycol[20];
		sscanf (sLine,  "%s %s %s %s %s",  gx, gy, gp, xrow, ycol);
		//cout<<"x: "<< gx <<" y: "<< gy << " " << gp <<" "<< xrow << " " << ycol <<endl;
		//for each node count the number of points within 
		// the search radius	
		int pointcnt = 0; //this is the cum number of points found
		for (int ii=0; ii<dataVec.size(); ii++) {
			float dx = dxvec[ii] - atof(gx);
			float dy = dyvec[ii] - atof(gy);
		 	
		 	float dx_dy = dx/dy;
			float rotation = atan(dx_dy);  //this is the rotaion angle from the x-axis
			float distance = sqrt(pow(dx, 2)+pow(dy, 2));  //simple distance function
			if (distance <= radius) {
        pointcnt++;
			}
		}
		///////////////////////////////////////////////////////////////////////////////
		//cout<< gx << " "<< gy << " " << pointcnt  << " " << xrow << " " << ycol <<endl;
		///////////////////////////////////////////////////////////////////////////////
		char sssLine[400];
		
		sscanf(gx,  "%s",  sssLine);
		strcat(sssLine, " ");
		strcat(sssLine, gy);
		strcat(sssLine, " ");
		char buf[255];
    sprintf(buf, "%i", pointcnt);
    strcat(sssLine, buf );
    strcat(sssLine, " ");
    strcat(sssLine, xrow);
    strcat(sssLine, " ");
    strcat(sssLine, ycol);
		out.push_back(sssLine);
	}
	return out;
}
