/* JUnit tests for CollectionActor class.
 *
 * Copyright (c) 2005 Natural Diversity Discovery Project.
 * All rights reserved.
 * 
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 * 
 * IN NO EVENT SHALL THE NATURAL DIVERSITY DISCOVERY PROJECT BE LIABLE 
 * TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR 
 * CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS SOFTWARE AND 
 * ITS DOCUMENTATION, EVEN IF THE NATURAL DIVERSITY DISCOVERY PROJECT
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE NATURAL DIVERSITY DISCOVERY PROJECT SPECIFICALLY DISCLAIMS ANY 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
 * PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE NATURAL 
 * DIVERSITY DISCOVERY PROJECT HAS NO OBLIGATION TO PROVIDE MAINTENANCE, 
 * SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.

   @ProposedRating Red (tmcphillips@naturaldiversity.org)
   @AcceptedRating Red (tmcphillips@naturaldiversity.org) 
 */
 
package org.nddp;

import org.nddp.exceptions.CollectionException;
import org.nddp.util.Parameters;
import ptolemy.data.expr.Parameter;
import ptolemy.kernel.CompositeEntity;
import ptolemy.kernel.util.IllegalActionException;
import ptolemy.kernel.util.NameDuplicationException;
import junit.framework.TestCase;

////////////////////////////////////////////////////////////////////////// 
//// TestCollectionActor 

/** 
 @author Timothy M. McPhillips
 */

public class TestCollectionActor  extends TestCase {

    Collection collection;
    CollectionActor_Local collectionActor;
    AtomicCoactor.TestProbe actorProbe;
	CollectionManager_Local collectionManager;
    CompositeEntity compositeEntity;
	
	
    public void setUp() throws Exception {
        
        collection = new Collection("collection");
        collectionManager = new CollectionManager_Local(
            collection, null, null);
        	compositeEntity = new CompositeEntity();
		collectionActor = 
		    new CollectionActor_Local(compositeEntity, "CollectionActorLocal");
		actorProbe = collectionActor.createTestProbe();
    }
    
	public void testConstructor() throws Exception {
	    
	    assertEquals(compositeEntity, collectionActor.getContainer());
	    assertEquals("CollectionActorLocal", collectionActor.getName());
	    
	    Parameter collectionPathParameter = 
	        (Parameter) collectionActor.getAttribute("collectionPath");
		assertEquals("collectionPath", collectionPathParameter.getName());
		assertEquals("", collectionPathParameter.getExpression());
	}
	
	public void testEnableParameterOverrideArray() throws Exception {
	    
	    Parameter parameterA = Parameters.stringParameter(collectionActor, "parameter A", "a");
	    Parameter parameterB = Parameters.stringParameter(collectionActor, "parameter B", "b");

	    assertEquals(0, actorProbe.overridableParametersMap().size());
	    collectionActor._enableParameterOverride(new Parameter[] {parameterA, parameterB});
	    assertEquals(2, actorProbe.overridableParametersMap().size());
	    assertTrue(actorProbe.overridableParametersMap().containsValue(parameterA));
	    assertTrue(actorProbe.overridableParametersMap().containsValue(parameterB));
	}

	public void testEnableParameterOverrideArray_NullArray() throws Exception {

	    Parameter[] nullParameterArray = null;
	    
	    try {
		    collectionActor._enableParameterOverride(nullParameterArray);
		    fail("A NullPointerException should have been thrown.");
	    }
	    catch(NullPointerException expected) {
	        assertEquals(
                "Null parameter array not allowed.",
                expected.getMessage());
	    }
	}
	
	public void testEnableParameterOverrideSingleton() throws Exception {
	    
	    Parameter parameterA = Parameters.stringParameter(collectionActor, "parameter A", "value");

	    assertEquals(0, actorProbe.overridableParametersMap().size());
	    collectionActor._enableParameterOverride(parameterA);
	    assertEquals(1, actorProbe.overridableParametersMap().size());
	    assertTrue(actorProbe.overridableParametersMap().containsValue(parameterA));
	}
	
	public void testEnableParameterOverrideSingleton_NullParameter() throws Exception {

	    Parameter nullParameter = null;
	    
	    try {
		    collectionActor._enableParameterOverride(nullParameter);
		    fail("A NullPointerException should have been thrown.");
	    }
	    catch(NullPointerException expected) {
	        assertEquals(
                "Null parameter not allowed.",
                expected.getMessage());
	    }
	}

	public void testEnableParameterOverrideSingleton_WrongContainer() throws Exception {

        AtomicCoactor actorOne = new AtomicCoactor(new CompositeEntity(), "Actor1");   
        AtomicCoactor actorTwo = new AtomicCoactor(new CompositeEntity(), "Actor2");   

	    Parameter parameterTwo = Parameters.stringParameter(actorTwo, "parameterTwo", "value");

	    try {
		    actorOne._enableParameterOverride(parameterTwo);
		    fail("An IllegalActionException should have been thrown.");
	    }
	    catch(IllegalActionException expected) {
	        assertEquals(
                "Parameter 'parameterTwo' is not contained by actor Actor1.",
                expected.getMessage());
	    }
	}

	public void testGetOuputPortNamed() throws Exception {

	    assertEquals(collectionActor.output1, 
	            collectionActor._getOutputPortNamed("output1"));

	    assertEquals(collectionActor.output2, 
	            collectionActor._getOutputPortNamed("output2"));
	}

	public void testGetOuputPortNamed_NoOutputPorts() throws Exception {

	    AtomicCoactor collectionActor = 
	        new AtomicCoactor(compositeEntity, "CollectionActor");

	    assertEquals(null, collectionActor._getOutputPortNamed("output1"));
	}

	public void testGetOuputPortNamed_NotOutputPort() throws Exception {

	    assertEquals(null, collectionActor._getOutputPortNamed("input1"));
	    assertEquals(null, collectionActor._getOutputPortNamed("notAPort"));
	}

	public void testGetOuputPortNamed_emptyPortName() throws Exception {

	    assertEquals(null, collectionActor._getOutputPortNamed(""));
	}
	
	public void testGetOuputPortNamed_nullPortName() throws Exception {

	    try {
		    collectionActor._getOutputPortNamed(null);
		    fail("A NullPointerException should have been thrown.");
	    }
	    catch(NullPointerException expected) {
	        assertEquals("Null portName not allowed.",
                expected.getMessage());
	    }
	}
	
//	public void testProcessCollectionException_Propagate() throws Exception {
//	    
//	    collectionActor.initialize();
//		collectionActor.propagateException = true;
//	    collectionActor.exceptionToThrow = 
//	        new CollectionException("error message");
//
//		collectionActor.handleOpeningDelimiter(collection.openingDelimiter());
//		
//		CollectionManager_Local collectionManager = (CollectionManager_Local)
//			actorProbe.collectionManager();
//	    
//	    	collectionActor.handleClosingDelimiter(collection.closingDelimiter());
//		
//	    assertEquals(collectionActor,
//            collectionManager.lastExceptionSource);
//	    assertEquals(collectionActor.exceptionToThrow,
//            collectionManager.lastException);
//	    assertEquals(true, collectionManager.ignored());
//	}
	
//	public void testProcessCollectionException_Discard() throws Exception {
//	    
//	    collectionActor.initialize();
//		collectionActor.propagateException = false;
//		collectionActor.exceptionToThrow = 
//		    new CollectionException("error message");
//
//		collectionActor.handleOpeningDelimiter(collection.openingDelimiter());
//		
//		CollectionManager_Local collectionManager = (CollectionManager_Local)
//			actorProbe.collectionManager();
//	    
//	    	collectionActor.handleClosingDelimiter(collection.closingDelimiter());
//		
//	    assertEquals(null, collectionManager.lastExceptionSource);
//	    assertEquals(null, collectionManager.lastException);
//	    assertEquals(false, collectionManager.ignored());
//	}
	
	private static class CollectionActor_Local extends AtomicCoactor {

    		public CollectionActor_Local(CompositeEntity container, String name) throws NameDuplicationException, IllegalActionException {
        		super(container, name);
        		
        		input1 = new CollectionPortBuffer();
        		input2 = CollectionIOPort.createInputPort(this, "input2");

        		output1 = CollectionIOPort.createOutputPort(this, "output1");
        		output2 = CollectionIOPort.createOutputPort(this, "output2");
        		
        		_setInputPort(input1);

        		attributeChanged(collectionPath);
    		}
    		
            public CollectionHandler.CollectionDisposition _handleCollectionStart(
   	        	CollectionManager collectionManager) 
	   		throws IllegalActionException, CollectionException {
   	    
   			return CollectionHandler.PROCESS_AND_FORWARD_COLLECTION;
   		}
   	
            public void _handleCollectionEnd(CollectionManager collectionManager) 
			throws IllegalActionException, CollectionException {

		    if (exceptionToThrow != null) {
	   	        throw exceptionToThrow;
	   	    }
	    }
	
		public IncomingCollectionManager manageIncomingCollection(
	    		Collection collection, 
	    		CollectionManager parentCollectionManager, 
	    		CollectionPort outputPort) 
	    		throws IllegalActionException {
		
			return new CollectionManager_Local(
		        collection, parentCollectionManager, outputPort);
		}
    		
        public CollectionHandler.ExceptionDisposition handleCollectionException(CollectionManager
	        collectionManager, CollectionException exception) {
	
			if (propagateException) {
			    return CollectionHandler.PROPAGATE_EXCEPTION;
			} else {
			    return CollectionHandler.DISCARD_EXCEPTION;
			}
		}
	
		public CollectionException exceptionToThrow = null;
    		public CollectionPortBuffer input1;
    		public boolean propagateException = false;
    		public CollectionPort input2;
    		public CollectionPort output1;
    		public CollectionPort output2;
	}

	
	private static class CollectionManager_Local 
		extends IncomingCollectionManager {

        	CollectionManager_Local(Collection collection, 
                CollectionManager parentCollectionManager, 
                CollectionPort outputPort) throws IllegalActionException {
            
            super(collection, parentCollectionManager, outputPort);
        	}
        
		public final void addException(AtomicCoactor source, 
	        CollectionException exception) {
		    
		    lastExceptionSource = source;
		    lastException = exception;
		}	

		private CollectionException lastException = null;
		private AtomicCoactor lastExceptionSource = null;
	}
}


